(:~
 : This module serves for getting all the XML serialization and its respective
 : preparation done.
 :
 : For displaying the XML data in HTML we first have to assemble all XML that
 : belong to a BdN work and then split it into smaller chunks (else each webpage
 : is too big to display it in a userfriendly way). The splitting is done by
 : considering two criteria:
 :
 : 1. which is the hierarchically highest tei:div of a work?
 : 2. does this tei:div have a tei:supplied[@type = "toc-title"] as descendant?
 :
 : ad 1: The hierarchical order of the tei:div is determined by its type: A part
 : may contain chapters and sections and a chapter may contain sections (part >
 : chapter > section). However, each work of the BdN has a slightly different
 : structure so we cannot simply assume that there is a tei:div[@type = "part"]
 : (some works may only have sections!).
 :
 : ad 2: For generating the navigation and generating HTML pages that make sense
 : in the context of a work, we also have to keep in mind the semantic structure
 : of a work that is indicated by tei:supplied[@type = "toc-title"]: Each unit
 : that has a tei:supplied[@type = "toc-title"] is a semantically coherent unit
 : that should be displayed together, thus should also be in an XML file together.
 :
 : Despite saving the whole and split XML files as well as their respective
 : intermediate format versions, we solve the whole serialization in memory in
 : order to avoid whitespace problems that we encounter when processing already
 : saved documents.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 :
 :)

xquery version "3.1";

module namespace xml-utils="http://bdn-edition.de/portal/xml-utils";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";
declare namespace html="http://www.w3.org/1999/html";
declare namespace xi="http://www.w3.org/2001/XInclude";
declare namespace textgrid="http://textgrid.info/namespaces/metadata/core/2010";
declare namespace bdn="http://www.bdn-edition.de/ns/1.0";

import module namespace bdn-utils="http://bdn-edition.de/ns/utils" at "utils.xqm";
import module namespace code-view="http://bdn-edition.de/ns/code-view";
import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace critical="http://bdn-edition.de/ns/critical" at "critical.xqm";
import module namespace digital="http://bdn-edition.de/ns/digital" at "digital.xqm";
import module namespace foreword="http://bdn-edition.de/ns/foreword" at "foreword.xqm";
import module namespace functx="http://www.functx.com";
import module namespace id-file="https://bdn-edition.de/ns/id-file" at "id-file.xqm";
(: DO NOT CALL dependent app with absolute path but rely on import resolution by namespace:)
import module namespace ifweb="http://bdn-edition.de/intermediate_format/ifweb" at "/db/apps/interformat/modules/intermediate_format/ifweb.xql";
import module namespace index-creation="http://bdn-edition.de/ns/index-creation" at "index-creation.xqm";
import module namespace sem-split="http://bdn-edition.de/portal/sem-split" at "sem-split.xqm";
import module namespace single="http://bdn-edition.de/ns/single" at "single.xqm";
import module namespace whitelist="https://bdn-edition.de/ns/whitelist" at "whitelist.xqm";
import module namespace internal-nav="http://bdn-edition.de/ns/internal-nav" at  "internal-nav.xqm";


declare option exist:serialize "method=xml media-type=text/xml expand-xincludes=no";

(:~
 : The main function creates a complete XML file for a BdN work and splits it
 : into smaller bits, up to chapter level. This is necessary since the single
 : XML files which are the basis for the complete file are ordered in chunks
 : that are not suitable for displaying them on a website.
 : Furthermore it prerenders all necessary HTML files.
 :
 : @author Michelle Weidling
 : @author Stefan Hynek
 : @version 1.1
 :
 :)
declare function xml-utils:main($author as xs:string) as xs:string*
{
    let $log := bdn-utils:create-log($author)
    let $init := bdn-utils:assure-dir-available("full-xmls")
    let $xml := xml-utils:make-xml($author, $log)
    let $xml := doc($config:full-xmls || $author || "-full.xml")
    return
    (
        internal-nav:main(),
        sem-split:main($author, $xml, $log),
        whitelist:main($author, $log),
        id-file:make-id-file($log),
        whitelist:add-nav-entries($author, $log),
        whitelist:mark-sv-on-chapter-level($author, $log),
        bdn-utils:add-log-entry($log, "#######################################################"),
        bdn-utils:add-log-entry($log, "### START HTML CREATION ###"),
        xml-utils:make-html($author, $log),
        index-creation:index-for-author($author, $log),
        bdn-utils:add-log-entry($log, "### FINISHED WHOLE PROCESS ###")
    )
};

(:~
 : For controlling and proof reading the data our scholars use the single files
 : that they create in TextGrid. Since all our transformations into the different
 : views base on the if, the single files have to have this pipe.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $filename The name of the file to be rendered, e.g. 12345.xml
 : @return strings containing the location of the stored HTML files
 :)
declare function xml-utils:main-single-file($filename as xs:string)
as xs:string*
{
    let $create-log :=
        if (doc-available($config:logs || "single-files-log.xml")) then
            ()
        else
            bdn-utils:create-log("single-files")
    let $log := concat($config:logs, "single-files-log.xml")
    let $uri := $config:sade-data || $filename


    let $assure-dir-available := bdn-utils:assure-dir-available("resources")
    let $assure-dir-available := bdn-utils:assure-dir-available("edition")
    let $assure-dir-available := bdn-utils:assure-dir-available("single-files")
    let $assure-dir-available := bdn-utils:assure-dir-available("single-files/xml")

    (: TODO: populate whitelist with files inside the aggregation and then
     : check against whitelisted files under <files />
     :)

    return
        if (not(
            (: structure files :)
            matches($filename, "25011") (: noesselt :)
            or matches($filename, "252bh") (: griesbach :)
            or matches($filename, "38bf6") (: less :)
            or matches($filename, "257w3") (: bahrdt-semler :)
            or matches($filename, "2skh7") (: teller :)
            or matches($filename, "3r960") (: steinbart :)
            or matches($filename, "3vpn2") (: sack :)
            (: various other files :)
            or matches($filename, "2r573") (: bibelstellen? :)
            (: published aggs, full xmls etc. :)
            or matches($filename, "3rkks") (: griesbach :)
            or matches($filename, "3rj87") (: griesbach :)
            or matches($filename, "3r96k") (: griesbach :)
            or matches($filename, "3rkkt") (: noesselt :)
            or matches($filename, "3rj88") (: noesselt :)
            or matches($filename, "234x3")) (: noesselt :)
        and matches($filename, ".xml")) then
            (: fails if structure file? :)
            let $author :=
                if (xmldb:get-mime-type($uri) = "text/xml") then
                    bdn-utils:get-author-from-whitelist($filename)
                else
                    ()

            let $if := ifweb:single-xml($filename, $config:sade-data, concat($config:single-files, "xml/"))

            return
                (
                    util:log("info", "make-html-single-file: " || $filename),
                    xml-utils:make-html-single-file($if, $filename, $log)
                    (: disabled because it triggers build of full-xml for each single file,
                     : even when uploaded as aggregation, see https://gitlab.gwdg.de/bibliothek-der-neologie/bdn-SADE/-/issues/204
                     : if ($author) then xml-utils:make-xml($author, $log) else ()
                     :)
                )
        else
            ()
};


(:~
 : Creates a complete XML file for a BdN work according to a given BdN author.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author The current author
 : @return the document path of the created XML file
 :)
declare function xml-utils:make-xml($author as xs:string, $log as xs:string)
as document-node()?
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Start XML creation")

    let $structure-file := bdn-utils:get-structure-file($author)
    let $filename := xml-utils:make-full-file-name($author)
    let $create-copy := xmldb:store($config:sade-data, $filename, $structure-file/tei:TEI)

    let $xincludes := $structure-file/tei:TEI//xi:include
    let $uris := $xincludes/@href
    let $resolved-file := doc($create-copy)

    let $update :=
    for $xinclude in $xincludes
      let $xi-filename := $xinclude/@href
      let $xpointer := $xinclude/@xpointer

      let $uri := substring-after($xi-filename, ':')

      let $source-doc := doc($config:sade-data || $uri || ".xml")
      let $source-node := $source-doc//*[@xml:id = $xpointer]

      let $node-to-be-replaced as node() := $resolved-file//*[@href = $xi-filename][@xpointer = $xpointer]
      return
        if ($source-node)
          then update replace $node-to-be-replaced with $source-node
          else ()

    let $store := xmldb:store($config:full-xmls, $filename, $resolved-file)

    return
        ($resolved-file,
        bdn-utils:add-log-entry($log, "Finished XML creation"))
};


(:~
 : Creates the intermediate format file for a BdN work according to a given BdN
 : author.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author The current author
 : @return The document path of the created XML files
 :)
declare function xml-utils:make-if ($author as xs:string) as xs:string?
{
  ifweb:main($author || "-full.xml")
};



(:~
 : Splits the current node, which can either be a tei:body of a volume or a
 : tei:div[@type = "part"] of a volume into its parts resp. chapters.
 : Since the splitting mechanism for creating parts reps. chapters is almost
 : identical, both are treated in this generic function.
 : In case there are any bigger nodes before the first tei:div[@type = "part"]
 : or tei:div[@type = "chapter"], these are stored in a separate file.
 :
 : Currently the only criterium to distinguish small and big nodes is whether
 : there is another tei:div before the first tei:div[@type = "part"] or
 : tei:div[@type = "chapter"]. This criterium may change in time, e.g.
 : depending on the author.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file-prefix A string containing information about the author, the
 : volume number, the part and the chapter
 : @param $current The current volume
 : @param $log The opened log file for the current author
 : @param $dir The directory in which the files are stored, e.g. "resources/edition/noesselt/xml"
 : @return The document path of the created XML file(s)
 :)
declare function xml-utils:split-to-part-or-chapter($file-prefix as xs:string,
$current as node(), $log as xs:string, $dir as xs:string) as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Split current " || name($current) || " " || $current/@type || " (" || $current/@xml:id || ") " || " to parts or chapters.")

    let $highest-child-type := xml-utils:get-highest-child-div-type($current)
    let $add-log-entry := bdn-utils:add-log-entry($log, "Highest child::tei:div/@type is " || $highest-child-type || ".")
    let $pre-unit-nodes := $current/child::tei:div[@type = $highest-child-type][1]/preceding-sibling::*
    let $pre-unit-nodes :=
        for $node in $pre-unit-nodes return
            (: section groups are surplus and only relevant for XIncludes :)
            if ($node[self::tei:div[@type = "section-group"]]) then
                $node/child::*
            else
                $node

    (: section groups are not included in $pre-unit-nodes, so we have to make sure they're not included.
    other tei:div of the types "introduction", "preface", ... :)
    let $children := $current/child::*[not(@type = "section-group")] except $pre-unit-nodes[not(self::tei:div)]
    let $children-no := count($children)
    let $add-log-entry := bdn-utils:add-log-entry($log, "Found " || $children-no || " child(ren).")
    for $iii in 1 to $children-no
        let $add-log-entry := bdn-utils:add-log-entry($log, "In child " || $iii || " (" || $children[$iii]/@xml:id ||").")
        let $content :=
          (: this covers tei:div of the types "introduction", "preface", ... in the pre-unit nodes :)
          if ($iii = 1 and $pre-unit-nodes[self::tei:div[not(@type = "section-group")]]) then
            ($children[$iii],
            bdn-utils:add-log-entry($log, "Pre-unit nodes stored separately, current content kept."))

          (: this covers the occurence of pre-unit nodes at all :)
          else if ($iii = 1 and $pre-unit-nodes) then
            (xml-utils:make-new-unit($children[$iii], $pre-unit-nodes),
            bdn-utils:add-log-entry($log, "New unit created by combining current content with pre-unit nodes."))

          else
            ($children[$iii],
            bdn-utils:add-log-entry($log, "No pre-unit nodes, content simply kept."))

        let $file-prefix := $file-prefix || "-" || $iii
        return
            if ($content and $iii = 1 and $pre-unit-nodes[self::tei:div]) then
              (
                  xml-utils:save-pre-unit-nodes($file-prefix || "-" || $iii, $pre-unit-nodes, $dir, $log),
                  xmldb:store($dir, $file-prefix || "-NIF.xml", $content),
                  ifweb:transform-split-xml($content, $file-prefix || ".xml", $dir),
                  bdn-utils:add-log-entry($log, "New pre-node unit including a tei:div saved as " || $file-prefix ||"-pre-unit.")
              )

          else
              if ($content and $highest-child-type = "part") then
              (
                bdn-utils:add-log-entry($log, "Current node is a tei:div[@type = 'part']. Continue splitting."),
                xml-utils:split-to-part-or-chapter($file-prefix, $content, $log, $dir)
              )

          else if ($content and $highest-child-type = ("chapter", "section")) then
            (
              xmldb:store($dir, $file-prefix || "-NIF.xml", $content),
              ifweb:transform-split-xml($content, $file-prefix || ".xml", $dir),
              bdn-utils:add-log-entry($log, "Part or chapter XML " || $file-prefix ||" created.")
            )

          else
            bdn-utils:add-log-entry($log, "An error occured while saving. Either $content
            was empty or there was no fitting @type.")
};

(:~
 : When there is a tei:div before a part, it is something like tei:div[@type
 : = "introduction"] and can be stored in a separate file since it most likely
 : will appear as a separate entry in the work's TOC. This function determines
 : if any nodes can be moved into a separate file.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $pre-unit-nodes All XML nodes that appear before the current unit
 : @result xs:boolean
 :)
declare function xml-utils:have-nodes-to-be-moved($pre-unit-nodes as node()*)
as xs:boolean
{
    if ($pre-unit-nodes and $pre-unit-nodes[self::tei:div]
    or $pre-unit-nodes and $pre-unit-nodes[self::ab[@type = "half-title"]]
    or not($pre-unit-nodes)) then
        false()
    else
        true()
};


(:~
 : Stores any nodes that appear before the current unit and that should be
 : moved.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file-prefix A string containing information about the author, chapter
 : etc., e.g. noesselt_1_2
 : @param $pre-unit-nodes All XML nodes that appear before the current unit
 : @param $move-pre-unit-nodes A Boolean flag
 : @result xs:string The location of the stored file
 :)
declare function xml-utils:save-pre-unit-nodes($file-prefix as xs:string,
$pre-unit-nodes as node()*, $dir as xs:string, $log as xs:string) as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Saving pre-unit nodes to " || $file-prefix || "-pre-unit-NIF.xml.")
    let $pre-unit-nodes :=
        if (count($pre-unit-nodes) gt 1) then
            element tei:div
            {
                attribute class {"split"},
                $pre-unit-nodes
            }
        else
            $pre-unit-nodes
    return
        (xmldb:store($dir, $file-prefix || "-pre-unit-NIF.xml", $pre-unit-nodes),
        ifweb:transform-split-xml($pre-unit-nodes, $file-prefix || "-pre-unit.xml", $dir))
};


(:~
 : Creates a new unit.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $unit The current unit
 : @param $pre-unit-nodes All XML nodes that appear before the current unit
 : @result element(tei:div) a tei:div containing the nodes before the current
 : unit and the unit content itself
 :)
declare function xml-utils:make-new-unit($unit as node(), $pre-unit-nodes as node()*)
as element(div)
{
    let $unit-contents := $unit/child::*
    return
        element {QName("http://www.tei-c.org/ns/1.0", "div")}
        {
            attribute type {$unit/@type},
            $pre-unit-nodes,
            $unit-contents
        }
};


(:~
 : Creates the full file name.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author The current author
 : @return The full file name
 :)
declare function xml-utils:make-full-file-name($author as xs:string) as xs:string
{
  $author || "-full.xml"
};


(:~
 : Retrieves the hierarchically highest type of all descending tei:div/@type.
 : The splitting of the full XML has to consider that all works of die BdN have
 : a slightly different structure.
 :
 : The hierarchical order of tei:div/@type is the following:
 : part > chapter > unit
 :
 : As additional criterium for splitting we have to mind which tei:div have
 : a tei:supplied[@type = "toc-title"], because every tei:div with such an
 : element as descendant has to appear in the navigation of the edition views.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $unit The current unit that is processed in xml-utils:split-to-part-or-chapter
 : @return The hierarchically highest type of all descending tei:div/@type
 :)
declare function xml-utils:get-highest-child-div-type($unit as node())
as xs:string
{
    let $possible-types := string-join(distinct-values($unit/child::tei:div/@type), " ")
    let $highest-type :=
            if (matches($possible-types, "part")) then
            "part"
        else if (matches($possible-types, "chapter")) then
            "chapter"
        else
            "section"
    return
        if ($unit/child::tei:div[@type = $highest-type]/tei:head[1]/descendant::tei:supplied[@reason = "toc-title"])
            then
                $highest-type
            else
                "section"
};


(:~
 : Creates an HTML file of the split XML files for each of our edition views
 : (critical, interactive and one for each printing).
 : TODO
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param
 : @return
 :)
declare function xml-utils:make-html($author as xs:string, $log as xs:string)
as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Start HTML creation")

    let $dir := bdn-utils:get-xml-dir($author)
    for $resource in collection($dir) return
        if (not(matches(base-uri($resource), "foreword"))) then
            let $filename := base-uri($resource) => functx:substring-after-last("/")
            return
                if (not(contains($filename, "git") or contains($filename, "NIF"))) then
                    (bdn-utils:add-log-entry($log, "Start HTML creation for " || $filename || "."),
                    if ($author = "bahrdt-semler") then
                        (xml-utils:make-html-digital-view($resource, $filename, $log),
                        xml-utils:make-html-code-view($resource, $filename, $log))
                    else
                        (
                            xml-utils:make-html-critical-view($resource, $filename, $log),
                            xml-utils:make-html-digital-view($resource, $filename, $log),
                            xml-utils:make-html-single-view($resource, $filename, $log),
                            xml-utils:make-html-code-view($resource, $filename, $log)
                        )
                    )
                else
                    ()
        else
            xml-utils:make-html-foreword($author, $log)
};


(:~
 : Creates for each single file an HTML file according to each of our edition
 : views (critical, digital and one for each printing).
 : TODO
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file The current file
 : @param $filename The file name, e.g. "2228v.xml"
 : @return The locations of the stored HTML files
 :)
declare function xml-utils:make-html-single-file($file as node()*,
$filename as xs:string, $log as xs:string) as xs:string*
{
  xml-utils:make-html-critical-view($file, $filename, $log),
  xml-utils:make-html-digital-view($file, $filename, $log),
  xml-utils:make-html-single-view($file, $filename, $log),
  xml-utils:make-html-code-view($file, $filename, $log)
};


(:~
 :
 : Creates an HTML serialization of the editorial foreword for a work.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author the BdN author's name
 : @param $log the path to the respective log file
 : @return The db location where the HTML is stored
 :
 :  :)
declare function xml-utils:make-html-foreword($author as xs:string, $log as xs:string)
as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Creating foreword for " || $author || ".")
    let $dir := $config:edition || $author || "/"

    let $foreword-xml := doc($dir || "xml/" || $author || "-foreword.xml")
    let $note-map := xml-utils:make-note-map($foreword-xml)
    let $foreword :=
        element xhtml:div
        {
            foreword:transform($foreword-xml, $note-map)
        }

    let $filename := $author || "-foreword.html"
    let $add-log-entry := bdn-utils:add-log-entry($log, "Successfully created " || $filename || ".")
    return
        xmldb:store($dir, $filename, $foreword)
};

(:~
 : Transforms a given file to a HTML file for the critical view.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file The current file in intermediate format
 : @param $filename
 : @return xs:string The path where the HTML is stored
 : :)
declare function xml-utils:make-html-critical-view($file as node()*,
$filename as xs:string*, $log as xs:string) as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Creating critical view for " || $filename || ".")

    let $rdg-map := xml-utils:make-rdg-map($file)
    let $content :=
        element xhtml:div
        {
            critical:transform($file, true(), $rdg-map)
        }
    let $new-filename := substring-before($filename, ".") || "-critical.html"
    let $data-dir := xml-utils:determine-dir($filename)
    let $add-log-entry := bdn-utils:add-log-entry($log, "Successfully created " || $new-filename || ".")
    return
        xmldb:store($data-dir, $new-filename, $content)
};


(:~
 : Transforms a given file to a HTML file for the single printing view for each
 : printing.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file The current file in intermediate format
 : @param $filename
 : @return xs:string The path where the HTML is stored
 : :)
declare function xml-utils:make-html-single-view($file as node()*,
$filename as xs:string*, $log as xs:string) as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Creating single views for " || $filename || ".")

(:    let $wits := "c":)
    let $wits :=
        (: single file names don't have a hyphen, only the ones that are the
            result of semantic splitting :)
        if (matches($filename, "-")) then
            let $author := bdn-utils:get-author-from-whitelist($filename)
            let $author := substring-before($filename, "-")
            let $full-file := doc($config:full-xmls || $author || "-full.xml")
            return $full-file//tei:witness/@xml:id
        else
            $file//tei:witness/@xml:id

    return
    for $wit in $wits
        let $content :=
            element xhtml:div
            {
                single:transform($file, $wit, $filename)
            }
        let $new-filename := substring-before($filename, ".") || "-" || $wit ||".html"
        let $data-dir := xml-utils:determine-dir($filename)
        let $add-log-entry := bdn-utils:add-log-entry($log, "Successfully created " || $new-filename || ".")

        return
            xmldb:store($data-dir, $new-filename, $content)
};


(:~
 : Transforms a given file to a HTML file for the digital view.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file The current file in intermediate format
 : @param $filename
 : @return xs:string The path where the HTML is stored
 : :)
declare function xml-utils:make-html-digital-view($file as node()*,
$filename as xs:string*, $log as xs:string) as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Creating digital view for " || $filename || ".")

    let $content := digital:transform($file, $filename)
    let $rdgs-and-quotes := xml-utils:find-rdgs-and-quotes($content)
    let $svg-paths :=
        element xhtml:div
        {
            $content//xhtml:path
        }
    let $content :=
        element xhtml:div{
            xml-utils:find-non-rdgs-and-non-quotes($content)
        }

    let $wits := distinct-values($rdgs-and-quotes/@wits[string-length(.) = 1])
    let $wits-panel := xml-utils:make-variants-overview($wits)
    let $variants-and-quotes :=
        element xhtml:div
        {
            $wits-panel,
            $rdgs-and-quotes
        }

    let $new-filename := substring-before($filename, ".") || "-digital.html"
    let $new-filename-variants := substring-before($filename, ".") || "-digital-variants-and-quotes.html"
    let $new-filename-paths := substring-before($filename, ".") || "-digital-svg-paths.html"
    let $data-dir := xml-utils:determine-dir($filename)
    let $add-log-entry := bdn-utils:add-log-entry($log, "Successfully created " || $new-filename || ", " || $new-filename-variants || ", " || $new-filename-paths || ".")

    return
        (xmldb:store($data-dir, $new-filename, $content),
        xmldb:store($data-dir, $new-filename-paths, $svg-paths),
        xmldb:store($data-dir, $new-filename-variants, $variants-and-quotes))
};


(:~
 : Returns all xhtml:span[@class = "rdg-digital"] that appear in a node set as a
 : result of digital:transform.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $content the result of digital:transform applied to a file
 : @return all xhtml:span[@class = "rdg-digital"]
 :)
declare function xml-utils:find-rdgs-and-quotes($content as node()*) as node()*
{
    ($content//*[matches(@class, "rdg-digital") and @wits],
    $content//*[matches(@class, "quoted-text")])
};


(:~
 : Returns all non-xhtml:span[@class = ("rdg-digital", "quoted-text")] that
 : appear in a node set as a result of digital:transform.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $content the result of digital:transform applied to a file
 : @return all xhtml:span[@class = "rdg-digital"]
 :)
declare function xml-utils:find-non-rdgs-and-non-quotes($content as node()*) as node()*
{
    for $node in $content return
    typeswitch ( $node )
      case text() return
          $node
      case comment() return
          $node
      default return
          if ($node[matches(@class, "rdg-digital")
          or matches(@class, "rdg-path")
          or matches(@class, "quoted-text")
          or matches(@class, "quote-path")]) then
              ()
          else
              element {$node/name()}
              {
                  $node/@*,
                  xml-utils:find-non-rdgs-and-non-quotes($node/node())
              }
};


declare function xml-utils:make-html-code-view($file as node()*,
$filename as xs:string*, $log as xs:string) as xs:string*
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Creating code view for " || $filename || ".")

    let $author :=
        if (matches($filename, "bahrdt-semler")) then
            "bahrdt-semler"
        else
            substring-before($filename, "-")
    (: take the non-IF file for display :)
    let $non-if-filename := substring-before($filename, ".") || "-NIF.xml"

    let $xml :=
        if ($author != "") then
            doc($config:edition || $author || "/xml/" || $non-if-filename)
        else
            doc($config:sade-data || $filename)

    let $serialized :=
        if ($author != "") then
            (: in split files the first element is sometimes a tei:div that serves
            as container but isn't encoded. this tei:div has an additional
            attribute, @class = "split", marking it as a wrapper :)
            let $content :=
                if ($xml/tei:div[not(@class = "split")] or $xml/tei:app[@type = "structural-variance"]) then
                    $xml/*
                (: when we have pre-unit nodes we sometimes have more than one
                element right before the next unit. in this case we have to keep
                the wrapping tei:div. :)
                else if (count($xml/tei:div/*) gt 1) then
                    $xml/*
                else
                    $xml/tei:div/*
            return
                code-view:main($content, true(), "hljs-", true())
        else
            (: select $xml//tei:TEI in order to exclude schema declarations :)
            code-view:main($xml/*, true(), "hljs-", true())

    let $content :=
        element xhtml:div
        {
            (: when we have more than one pre-unit node, we have kept the wrapper
            tei:div[@class = "split"] for serializing. it shouldn't be displayed
            in the code view, though. :)
            if ($serialized[self::tei:div and @class = "split"]) then
                $serialized/*
            else
                $serialized
        }
    let $new-filename := substring-before($filename, ".") || "-code.html"
    let $data-dir := xml-utils:determine-dir($filename)
    let $add-log-entry := bdn-utils:add-log-entry($log, "Successfully created " || $new-filename || ".")

    return xmldb:store($data-dir, $new-filename, $content)
};


(:~
 : Determines the directory where an HTML file should be saved.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $filename the name of the current file, e.g. "2228v.xml" or "griesbach-1-1.xml
 : @return xs:string The path where the HTML is stored
 :   :)
declare function xml-utils:determine-dir($filename as xs:string) as xs:string
{
    if (matches($filename, "-")) then
        let $author :=
            if (not(matches($filename, "bahrdt-semler"))) then
                substring-before($filename, "-")
            else
                "bahrdt-semler"
        return
            $config:edition || $author

    else
        let $assure-dir-available := bdn-utils:assure-dir-available("single-files")
        return
            $config:edition || "single-files/"
};

(:~
 : In the critical view we provide a section bound counting of the tei:rdg that
 : appear in the critical apparatus (which are tei:rdg[@type = ("v", "pp", "pt")])).
 :
 : This function provides map with the IDs of all tei:div[@type] and tei:head outside section elements of a resource, their
 : respective relevant tei:rdg and the tei:rdg's position within the tei:div or tei:head.
 :
 : The map's design is:
 :
 : map
 : {
 :  "element-id":
 :      map
 :      {
 :        "rdg-id": rdg-number-in-the-element
 :      }
 : }
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $resource an XML file that's the result of the splitting process, e.g. doc("/db/apps/sade/resources/edition/noesselt/xml/noesselt-1-front.xml")
 : @return map(*)* a map containing information about the rdg number/position within a tei:div (cf. description)
 :)
declare function xml-utils:make-rdg-map($resource as node()*) as map(*)*
{
    let $elements :=
    (
        $resource//*[self::tei:div[@type][@type != "section-group"] or self::tei:app[@type = "standalone"]],
        $resource/tei:div[@class = 'split']/child::tei:head,
        $resource//tei:head[not(ancestor::tei:div[@type = "section"])]
    )

    return
      map:merge
      (
        for $element in $elements
          return map:entry
          (
            $element/@id,
            map:merge
            (
              for $rdg in $element//tei:rdg[@type = ("v", "pp", "pt")]
                let $position := count($element/descendant::tei:rdg[@type = ("v", "pp", "pt")][. << $rdg]) + 1
                return map:entry($rdg/@id, $position)
            )
          )
      )
};

(:~
 : Creates an overview of which printing appear in a given text. A click on the
 : generated button for a printing displays all available rdgs of this printing
 : in div[@id = "variants"].
 :
 : This function is called by xml-utils:make-html-digital-view.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 : @param $wits a string array of wits that appear in the text
 : @return xhtml:div containing a button for each readings
 :)
declare function xml-utils:make-variants-overview($wits as xs:string*)
as element(div)
{
  <div class="text-center">
    <div class="btn-group variants-btns" role="group" aria-label="Basic example">
    {
      for $wit in $wits order by $wit
      let $id := "variants-btn-" || $wit
      return
        <button type="button" id="{$id}" class="btn btn-secondary variants-btn">{$wit}</button>
    }
    </div>
  </div>
};


(:~
 : An auxiliary function that maps the tei:note IDs of a modern
 : editor's foreword to a md5 hash.
 :
 : This function is called by xml-utils:make-html-foreword.
 :
 : @author Michelle Weidling
 : @param $foreword An opened foreword XML
 : @return A map for tei:note's IDs and their hashes
 : :)
declare function xml-utils:make-note-map($foreword as node()*) as map(*)
{
  map:merge
  (
    for $note in $foreword//tei:note
    return
      map:entry( string($note/@id), util:hash($note/@id, "md5") )
  )
};
